/* Name: main.c
 * Project: Gamecube/N64 to USB converter
 * Author: Raphael Assenat <raph@raphnet.net
 * Copyright: (C) 2007-2011 Raphael Assenat <raph@raphnet.net>
 * License: Proprietary, free under certain conditions. See Documentation.
 * Comments: Based on HID-Test by Christian Starkjohann
 * Tabsize: 4
 */
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/sleep.h>
#include <avr/wdt.h>
#include <util/delay.h>
#include <string.h>

#include "usbdrv.h"
#include "oddebug.h"
#include "gamepad.h"
#include "gamecube.h"
#include "n64.h"

#include "devdesc.h"
#include "reportdesc.h"

static uchar *rt_usbHidReportDescriptor=NULL;
static uchar rt_usbHidReportDescriptorSize=0;
static uchar *rt_usbDeviceDescriptor=NULL;
static uchar rt_usbDeviceDescriptorSize=0;


int usbCfgSerialNumberStringDescriptor[] PROGMEM = {
 	USB_STRING_DESCRIPTOR_HEADER(USB_CFG_SERIAL_NUMBER_LENGTH),
 	'1', '0', '0', '0'
 };

static Gamepad *curGamepad = NULL;


/* ----------------------- hardware I/O abstraction ------------------------ */

static void hardwareInit(void)
{
	// init port C as input with pullup
	DDRC = 0x00;
	PORTC = 0xff;
	
	/* 1101 1000 bin: activate pull-ups except on USB lines 
	 *
	 * USB signals are on bit 0 and 2. 
	 *
	 * Bit 1 is connected with bit 0 (rev.C pcb error), so the pullup
	 * is not enabled.
	 * */
	PORTD = 0xf8;   

	/* Usb pin are init as outputs */
	DDRD = 0x01 | 0x04;    
	_delay_ms(11); 
	DDRD = 0x00;    /* 0000 0000 bin: remove USB reset condition */
			/* configure timer 0 for a rate of 12M/(1024 * 256) = 45.78 Hz (~22ms) */
	TCCR0 = 5;      /* timer 0 prescaler: 1024 */

	TCCR2 = (1<<WGM21)|(1<<CS22)|(1<<CS21)|(1<<CS20);
	OCR2 = 196; // for 60 hz

}

static uchar    reportBuffer[8];    /* buffer for HID reports */



/* ------------------------------------------------------------------------- */
/* ----------------------------- USB interface ----------------------------- */
/* ------------------------------------------------------------------------- */

static uchar    idleRate;           /* in 4 ms units */

uchar   usbFunctionDescriptor(struct usbRequest *rq)
{
	if ((rq->bmRequestType & USBRQ_TYPE_MASK) != USBRQ_TYPE_STANDARD)
		return 0;

	if (rq->bRequest == USBRQ_GET_DESCRIPTOR)
	{
		// USB spec 9.4.3, high byte is descriptor type
		switch (rq->wValue.bytes[1])
		{
			case USBDESCR_DEVICE:
				usbMsgPtr = rt_usbDeviceDescriptor;
				return rt_usbDeviceDescriptorSize;

			case USBDESCR_HID_REPORT:
				usbMsgPtr = rt_usbHidReportDescriptor;
				return rt_usbHidReportDescriptorSize;

		}
	}

	return 0;
}

static int getGamepadReport(unsigned char *dstbuf)
{
	if (curGamepad == NULL) {
		dstbuf[0] = 0x7f;
		dstbuf[1] = 0x7f;
		dstbuf[2] = 0x7f;
		dstbuf[3] = 0x7f;
		dstbuf[4] = 0x7f;
		dstbuf[5] = 0x7f;
		dstbuf[6] = 0;
		dstbuf[7] = 0;
		return 8;
	}
	else {
		curGamepad->buildReport(dstbuf);
		return curGamepad->report_size;
	}
}


uchar	usbFunctionSetup(uchar data[8])
{
	usbRequest_t    *rq = (void *)data;

	usbMsgPtr = reportBuffer;
	if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS){    /* class request type */
		if(rq->bRequest == USBRQ_HID_GET_REPORT){  /* wValue: ReportType (highbyte), ReportID (lowbyte) */
			/* we only have one report type, so don't look at wValue */
			//curGamepad->buildReport(reportBuffer);
			//return curGamepad->report_size;
			return getGamepadReport(reportBuffer);
		}else if(rq->bRequest == USBRQ_HID_GET_IDLE){
			usbMsgPtr = &idleRate;
			return 1;
		}else if(rq->bRequest == USBRQ_HID_SET_IDLE){
			idleRate = rq->wValue.bytes[1];
		}
	}else{
	/* no vendor specific requests implemented */
	}
	return 0;
}

/* ------------------------------------------------------------------------- */

int main(void)
{
	char must_report = 0, first_run = 1;
	uchar   idleCounter = 0;
	unsigned char state;
	int error_count = 0;

	hardwareInit();
	odDebugInit();

#if 0
	while(1)
	{
		/* Check for gamecube controller */
		curGamepad = gamecubeGetGamepad();
		curGamepad->init();
		if (curGamepad->probe()) 
			break;

		_delay_ms(30);

		/* Check for n64 controller */
		curGamepad = n64GetGamepad();
		curGamepad->init();
		if (curGamepad->probe()) 
			break;
		
		_delay_ms(30);
	}
#endif

	/* I know both gamepads uses the same report descriptor and
	 * the same device descriptor */
	rt_usbHidReportDescriptor = gcn64_usbHidReportDescriptor;
	rt_usbHidReportDescriptorSize = getUsbHidReportDescriptor_size();
	rt_usbDeviceDescriptor = (void*)usbDescrDevice;
	rt_usbDeviceDescriptorSize = getUsbDescrDevice_size();

	// Do hardwareInit again. It causes a USB reset. 
	hardwareInit();
//	curGamepad->init();

	set_sleep_mode(SLEEP_MODE_IDLE);
	wdt_enable(WDTO_2S);
	usbInit();

	sei();
	DBG1(0x00, 0, 0);
	
	
	while (1) 
	{	
		wdt_reset();

	
		if (error_count > 30) {
			curGamepad = NULL;
		}


		if (curGamepad == NULL) {
			// this must be called at each 50 ms or less
			usbPoll();

			/* Check for gamecube controller */
			curGamepad = gamecubeGetGamepad();
			curGamepad->init();
			if (curGamepad->probe()) {
				error_count = 0;
				goto detected;
			}

			_delay_ms(40);
			usbPoll();

			/* Check for n64 controller */
			curGamepad = n64GetGamepad();
			curGamepad->init();
			if (curGamepad->probe())  {
				error_count = 0;
				goto detected;
			}
			
			_delay_ms(40);
			usbPoll();

			continue;
		}
detected:
		
		/* main event loop */
		wdt_reset();

		// this must be called at each 50 ms or less
		usbPoll();
		if (first_run) {
			if (curGamepad->update()) {
				error_count++;
			} else {
				error_count = 0;
			}
			first_run = 0;
		}

		if(TIFR & (1<<TOV0)){   /* 22 ms timer */
			TIFR = 1<<TOV0;
			if(idleRate != 0){
				if(idleCounter > 4){
					idleCounter -= 5;   /* 22 ms in units of 4 ms */
				}else{
					idleCounter = idleRate;
					must_report = 1;
				}
			}
		}

		if (TIFR & (1<<OCF2))
		{
			TIFR = 1<<OCF2;
			if (!must_report)
			{
				// Wait! Before doing this, let an USB interrupt occur. This
				// prevents USB interrupts from occuring during the
				// timing sensitive Gamecube/N64 communication
				//
				// USB communication interrupts are triggering at regular
				// intervals on my machine. Between interrupts, we have 900uS of
				// free time.
				//
				// The trick here is to put the CPU in idle mode ; That is, wating
				// for interrupts, doing nothing. When the CPU resumes, an interrupt
				// has been serviced. The final delay helps when we get more than
				// once in a row (it happens, saw it on the scope. It was inserting
				// a huge delay in the command I was sending to the controller)
				//
				sleep_enable();	
				sleep_cpu();
				sleep_disable();
				_delay_us(100);

				if (curGamepad->update()) {
					error_count++;
				} else {
					error_count = 0;
				}

				if (curGamepad->changed()) {
					must_report = 1;
				}
			}

		}
		
			
		if(must_report)
		{
			if (usbInterruptIsReady())
			{ 	
				must_report = 0;

				curGamepad->buildReport(reportBuffer);
				usbSetInterrupt(reportBuffer, curGamepad->report_size);
			}
		}
	}
	return 0;
}

/* ------------------------------------------------------------------------- */
