/* Name: main.c
 * Project: Gamecube/N64 to USB converter
 * Author: Raphael Assenat <raph@raphnet.net
 * Copyright: (C) 2007-2011 Raphael Assenat <raph@raphnet.net>
 * License: Proprietary, free under certain conditions. See Documentation.
 * Tabsize: 4
 */
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <string.h>
#include "gamepad.h"
#include "leds.h"
#include "n64.h"
#include "reportdesc.h"


/******** IO port definitions **************/
#define N64_DATA_PORT	PORTC
#define N64_DATA_DDR	DDRC
#define N64_DATA_PIN	PINC
#define N64_DATA_BIT	(1<<5)

/*********** prototypes *************/
static void n64Init(void);
static char n64Update(void);
static char n64Changed(void);
static void n64BuildReport(unsigned char *reportBuffer);


/* What was most recently read from the controller */
static unsigned char last_built_report[GCN64_REPORT_SIZE];

/* What was most recently sent to the host */
static unsigned char last_sent_report[GCN64_REPORT_SIZE];

static void n64Init(void)
{
	unsigned char sreg;
	sreg = SREG;
	cli();

	DDRB |= 0x02; // Bit 1 out
	PORTB &= ~0x02; // 0
	
	// data as input
	N64_DATA_DDR &= ~(N64_DATA_BIT);

	// keep data low. By toggling the direction, we make the
	// pin act as an open-drain output.
	N64_DATA_PORT &= ~N64_DATA_BIT;

	SREG = sreg;

	n64Update();

}

static void n64_sendCmd(unsigned char cmd)
{
	/*
	 * r16: Holds the bit to be AND'ed with current byte
	 * r17: Holds the current byte
	 * r18: Loop counter for delays
	 *
	 * Edit sbi/cbi/andi instructions to use the right bit!
	 * 
	 * 3 us == 36 cycles
	 * 1 us == 12 cycles
	 */
	asm volatile(		
			
"			ldi r16, 0x80			\n" // 1
"nextBit:							\n" 
"			mov r17, %3				\n" // 2
//"			ldi r17, 0x0f			\n" // 2
"			and r17, r16			\n" // 1
"			breq send0				\n" // 2
"			nop						\n"

"send1:								\n"
"			sbi %0, 5				\n" // 2
"				nop\nnop\nnop\nnop	\n" // 4
"				nop\nnop\nnop\nnop	\n" // 4
"				nop\nnop\n			\n" // 3
"			cbi %0, 5				\n" // 2
"				ldi r18, 8			\n"	// 1 
"lp1:			dec r18				\n"	// 1 
"				brne lp1			\n"	// 2
"				nop\nnop			\n" // 2
"			lsr r16					\n" // 1
"			breq done				\n" // 1
"			rjmp nextBit			\n" // 2
		
/* Send a 0: 3us Low, 1us High */
"send0:		sbi %0, 5				\n"	// 2
"				ldi r18, 11			\n"	// 1
"lp0:			dec r18				\n"	// 1
"				brne lp0			\n"	// 2
"				nop					\n" // 1

"          	cbi %0,5				\n" // 2
"				nop\nnop\n				\n" // 2

"			lsr r16					\n" // 1
"			breq done				\n" // 1
"			rjmp nextBit			\n" // 2

"done:								\n"
"			cbi %0, 5				\n"
"			nop\nnop\nnop\nnop		\n"	// 4
"			nop\nnop\nnop\nnop		\n"	// 4
"			nop						\n"	// 1


// Stop bit (1us low, 3us high)
"          	sbi %0,5				\n" // 2
"			nop\nnop\nnop\nnop		\n" // 4
"			cbi %0,5				\n" 

// For the stop bit high part, just let it go and wait for
// the signal to become high.

"			ldi r16, 0xff\n			" // setup a timeout
"waitHigh:\n						"
"			dec r16\n				" // decrement timeout
"			breq finish_wait_high\n				" // handle timeout condition
"			sbis %1, 5\n			" // Read the port
"			rjmp waitHigh\n			"

"finish_wait_high:"

			: 	
			: 	"I" (_SFR_IO_ADDR(N64_DATA_DDR)), // %0
				"I" (_SFR_IO_ADDR(N64_DATA_PIN)),	// %1
				"I" (_SFR_IO_ADDR(PORTB)),			// %2
				"r"(cmd) 							// %3
			: "r16","r17","r18"
			);
}


/**
 * \brief Receive N64/GC style data
 *
 * \param dstbuf Destination buffer
 * \return The number of received bytes. 0 means nothing or overflow.
 *
 * This function times the low and high states
 * of the N64/GC bus until the buffer is filled
 * with level timings or timeout.
 */
static unsigned char n64_receive(unsigned char dstbuf[256])
{
	unsigned char count;
	
	// The data line has been released. 
	// The receive part below expects it to be still high
	// and will wait for it to become low before beginning
	// the counting.
	asm volatile(
		"	push r30				\n"	// save Z
		"	push r31				\n"	// save Z

//		"	sbi %3,1				\n" // DEBUG

		"	clr %0					\n"
		"	clr r16					\n"
"initial_wait_low:\n"
		"	inc r16					\n"
		"	breq timeout			\n" // overflow to 0
		"	sbic %2, 5				\n"
		"	rjmp initial_wait_low	\n"

		"	cbi %3, 1				\n" // DEBUG
	
		// the next transition is to a high bit	
		"	rjmp waithigh			\n"

"waitlow:\n"
		"	clr r16					\n"
"waitlow_lp:\n"
		"	inc r16					\n"
		"	breq timeout			\n" // overflow to 0
		"	sbic %2, 5				\n"
		"	rjmp waitlow_lp			\n"
	
		"	inc %0					\n" // count this timed low level
		"	breq overflow			\n"
		"	st z+,r16				\n"

"waithigh:\n"
		"	clr r16					\n"
"waithigh_lp:\n"
		"	inc r16					\n"
		"	breq timeout			\n" // overflow to 0
		"	sbis %2, 5				\n"
		"	rjmp waithigh_lp		\n"
	
		"	inc %0					\n" // count this timed high level
		"	breq overflow			\n"
		"	st z+,r16				\n"

		"	rjmp waitlow			\n"

"overflow:  \n"
		"	sbi %3,1				\n" // DEBUG
"timeout:	\n"
"			pop r31				\n" // restore z
"			pop r30				\n" // restore z

		: 	"=r" (count)						// %0
		: 	"z"((volatile void*)dstbuf),		// %1
			"I" (_SFR_IO_ADDR(N64_DATA_PIN)),	// %2
			"I" (_SFR_IO_ADDR(PORTB))			// %3
		: 	"r16"
	);

	return count;
}

int _n64Update(unsigned char tmp)
{
	int i;
	unsigned char tmpdata[4];	
	unsigned char count;
	unsigned char results[256];

	n64_sendCmd(tmp);

	count = n64_receive(results);

	if (count != 65) {
		return -1;
	}

//	PORTB |= 2;
//	PORTB &= ~2;

	//
	//          ________
	// ________/
	//
	//   [i*2]    [i*2+1]
	//
	//          ________________
	// 0 : ____/
	//                      ____
	// 1 : ________________/
	//
	// The timings on a real N64 are
	//  
	// 0 : 1 us low, 3 us high
	// 1 : 3 us low, 1 us high
	//
	// However, HORI pads use something similar to
	//
	// 0 : 1.5 us low, 4.5 us high
	// 1 : 4.5 us low, 1.5 us high
	//
	//
	// No64 us = microseconds
	//
	//

	for (i=0; i<count/2; i++)
	{
		results[i] = 
			results[i*2] < results[i*2+1] ? 1 : 0;

	}


/*
	Bit	Function
	0	A
	1	B
	2	Z
	3	Start
	4	Directional Up
	5	Directional Down
	6	Directional Left
	7	Directional Right
	8	unknown (always 0)
	9	unknown (always 0)
	10	L
	11	R
	12	C Up
	13	C Down
	14	C Left
	15	C Right
	16-23: analog X axis
	24-31: analog Y axis
 */

	tmpdata[0]=0;
	tmpdata[1]=0;
	tmpdata[2]=0;
	tmpdata[3]=0;

	for (i=0; i<4; i++) // A B Z START
		tmpdata[2] |= results[i] ? (0x01<<i) : 0;

	for (i=0; i<4; i++) // C-UP C-DOWN C-LEFT C-RIGHT
		tmpdata[2] |= results[i+12] ? (0x10<<i) : 0;
	
	for (i=0; i<2; i++) // L R
		tmpdata[3] |= results[i+10] ? (0x01<<i) : 0;

	for (i=0; i<8; i++) // X axis
		tmpdata[0] |= results[i+16] ? (0x80>>i) : 0;
	
	for (i=0; i<8; i++) // Y axis
		tmpdata[1] |= results[i+24] ? (0x80>>i) : 0;	

	// analog joystick
	last_built_report[0] = ((int)((signed char)tmpdata[0]))+127;
	last_built_report[1] = ((int)( -((signed char)tmpdata[1])) )+127;

	last_built_report[2] = 0x7f;
	last_built_report[3] = 0x7f;
	last_built_report[4] = 0x7f;
	last_built_report[5] = 0x7f;

	// buttons
	last_built_report[6] = tmpdata[2];
	last_built_report[7] = tmpdata[3];

	// dpad as buttons
	if (results[4]) 
		last_built_report[7] |= 0x04;
	if (results[5])
		last_built_report[7] |= 0x08;
	if (results[6])
		last_built_report[7] |= 0x10;
	if (results[7])
		last_built_report[7] |= 0x20;

	return 0;
}

static char n64Update(void)
{
	return _n64Update(0x01); // get status
}

static char n64Probe(void)
{
	char i;

	for (i=0; i<15; i++)
	{
		_delay_ms(30);
		
		if (0 == n64Update()) {
			return 1;
		}
	}
	return 0;
}


static char n64Changed(void)
{
	static int first = 1;
	if (first) { first = 0;  return 1; }
	
	return memcmp(last_built_report, last_sent_report, GCN64_REPORT_SIZE);
}

static void n64BuildReport(unsigned char *reportBuffer)
{
	if (reportBuffer) 
		memcpy(reportBuffer, last_built_report, GCN64_REPORT_SIZE);
	
	memcpy(	last_sent_report, last_built_report, GCN64_REPORT_SIZE);	
}

Gamepad N64Gamepad = {
	.report_size			= GCN64_REPORT_SIZE,
	.init					= n64Init,
	.update					= n64Update,
	.changed				= n64Changed,
	.buildReport			= n64BuildReport,
	.probe					= n64Probe,
};

Gamepad *n64GetGamepad(void)
{
	N64Gamepad.reportDescriptor = (void*)gcn64_usbHidReportDescriptor;
	N64Gamepad.reportDescriptorSize = getUsbHidReportDescriptor_size();
	return &N64Gamepad;
}

